/*

EEPROM layout

Byte
0       NodeID
1       RF Freq
2       Network Group
3-6     vCal
7-10    i1Cal
11-14   i1Lead
15-18   i2Cal
19-22   i2Lead
23-26   i3Cal
27-30   i3Lead
31-34   i4Cal
35-38   i4Lead
39-42   Datalogging period
43      Pulses enabled
44-45   Pulse min period
46      Temperatures enabled
47-94   Temperature Addresses (6×8)
95      Data whitening
96      RF Power
97      showCurrents print enabled
98-101  vCal_USA
*/


#include <Arduino.h>
#include <avr/pgmspace.h>
#include <EEPROM.h>

 // Available Serial Commands
const PROGMEM char helpText1[] =                                
"\n"
"Available commands for config during start-up:\n"
"  b<n>      - set r.f. band n = a single numeral: 4 = 433MHz, 8 = 868MHz, 9 = 915MHz (may require hardware change)\n"
"  f<x>      - x = the line frequency in Hz: normally either 50 or 60\n"
"  g<nnn>    - set Network Group  nnn - an integer (OEM default = 210)\n"
"  k<n>      - turn on-line calibration on\n"
"  i<nn>     - set node ID nn = an integer (standard node ids are 1..30)\n"
"  p<nn>     - set RF Power nn = an integer, 0 - 31. 0 = -18 dBm, 31 = +13 dBm, default = 25 (+7 dBm)\n"
"                 DO NOT USE a value > 25 unless an antenna is fitted.\n"
"  r         - restore sketch defaults\n"
"  s         - save config to EEPROM\n"
"  v         - show firmware version\n"
"  w<n>      - turn RFM Wireless data on or off:\n"
"                 n = 0 for OFF, n = 1 for ON, n = 2 for ON with whitening\n"
"  x         - exit and continue\n"
"  ?         - show this text again\n"
"\n"
"Available commands only when running:\n"
"  c<n>      - enable voltage, current & power factor values to serial output for calibration. n = 0 for OFF, n = 1 for ON,\n"
"  k<x> <yy.y> <zz.z>\n"
"            - Calibrate an analogue input channel:\n"
"                 x = a single numeral: 0 = voltage calibration, 1 = ct1 calibration, 2 = ct2 calibration, etc\n"
"                 yy.y = a floating point number for the voltage/current calibration constant\n"
"                 zz.z = a floating point number for the phase calibration for this c.t. (z is not needed, or ignored if supplied, when x = 0)\n"
"                 e.g. k0 256.8\n"
"                      k1 90.9 2.00\n"
"  l         - list the config values\n"
"  m<x> <yy> - meter pulse counting:\n"
"                 x = 0 for OFF, x = 1 for ON, <yy> = an integer for the pulse minimum period in ms. (y is not needed, or ignored when x = 0)\n"
"  p<xx.x>   - xx.x = a floating point number for the datalogging period\n" 
"  s         - save config to EEPROM\n"
"  t0 <y>    - turn temperature measurement on or off:\n"
"                 y = 0 for OFF, y = 1 for ON\n"
"  t<x> <yy> <yy> <yy> <yy> <yy> <yy> <yy> <yy>\n"
"                 change a temperature sensor's address or position:\n"
"                 x = a single numeral: the position of the sensor in the list (1-based)\n"
"                 yy = 8 hexadecimal bytes representing the sensor's address\n"
"                   e.g.  28 81 43 31 07 00 00 D9\n"
"                   N.B. Sensors CANNOT be added.\n"
"  ?         - show this text again\n"
;


extern DeviceAddress *temperatureSensors;

static void load_config(bool verbose)
{
  unsigned int dataPresent;
 
  if (dataPresent = eepromRead(eepromSig, (byte *)&EEProm))
  {
    if (verbose)
    {
      if (dataPresent) {
        Serial.println(F("Loaded EEPROM config"));
      } else { 
        Serial.println(F("No EEPROM config"));
      }
      list_calibration();
    }
  }
}

static void list_calibration(void)
{
  
  Serial.println(F("Settings:"));
  Serial.print(F("Group ")); Serial.print(EEProm.networkGroup);
  Serial.print(F(", Node ")); Serial.print(EEProm.nodeID & 0x1F);
  Serial.print(F(", Band ")); 
  Serial.print(EEProm.RF_freq == RF12_433MHZ ? 433 : 
               EEProm.RF_freq == RF12_868MHZ ? 868 :
               EEProm.RF_freq == RF12_915MHZ ? 915 : 0);
  Serial.print(F(" MHz\n\n"));
 
  Serial.println(F("Calibration:"));
  Serial.print(F("vCal = ")); Serial.println(USA?EEProm.vCal_USA:EEProm.vCal);
  Serial.print(F("i1Cal = ")); Serial.println(EEProm.i1Cal);
  Serial.print(F("i1Lead = ")); Serial.println(EEProm.i1Lead);
  Serial.print(F("i2Cal = ")); Serial.println(EEProm.i2Cal);
  Serial.print(F("i2Lead = ")); Serial.println(EEProm.i2Lead);
  Serial.print(F("i3Cal = ")); Serial.println(EEProm.i3Cal);
  Serial.print(F("i3Lead = ")); Serial.println(EEProm.i3Lead);
  Serial.print(F("i4Cal = ")); Serial.println(EEProm.i4Cal);
  Serial.print(F("i4Lead = ")); Serial.println(EEProm.i4Lead);
  Serial.print(F("datalog = ")); Serial.println(EEProm.period);
  Serial.print(F("pulses = ")); Serial.println(EEProm.pulse_enable);
  Serial.print(F("pulse period = ")); Serial.println(EEProm.pulse_period);
  Serial.print(F("temp_enable = ")); Serial.println(EEProm.temp_enable);
  printTemperatureSensorAddresses();
  Serial.print(EEProm.rf_whitening ? (EEProm.rf_whitening ==1 ? "RF on":"RF whitened"):"RF off"); Serial.print("\n");
  Serial.print(F("RF power (dBm) = ")); Serial.println((EEProm.rf_power & 0x1F) - 18);

}

static void save_config()
{
  Serial.println("Saving...");
  eepromWrite(eepromSig, (byte *)&EEProm, sizeof(EEProm));
  eepromPrint();
  Serial.println(F("\n\nDone. New config saved to EEPROM"));
}

static void wipe_eeprom(void)
{
  Serial.println(F("Resetting..."));
  
  eepromHide(eepromSig);   
  Serial.println(F("Done. Sketch will now restart using default config."));
  delay(200);
}

void softReset(void)
{
  asm volatile ("  jmp 0");
}

void readConfigInput(void)
{
  Serial.println(F("POST.....wait 10s"));
  Serial.println(F("'+++' then [Enter] for config mode"));
    
  unsigned long start = millis();
  bool done = false;
  while (millis() < (start + 10000))
  {
    // If serial input of keyword string '+++' is entered during 10s POST then enter config mode
    if (Serial.available())
      if ( Serial.readString() == "+++\r\n")
      {
        Serial.println(F("Entering config mode..."));
        showString(helpText1);
        while(config())
          ;
      }
  }
}


static bool config(void) 
{
/*
 * Reads calibration information (if available an permitted) from the serial port at startup
 *  see the user instruction above, the comments below or the separate documentation for details
 */ 
  if (Serial.available())
  {
    char c = Serial.peek();
    int k1;
    switch (c) {
        case 'i':  //  Set NodeID )range expected: 1 - 30
          EEProm.nodeID = Serial.parseFloat();
          Serial.print(F("Node ")); Serial.println(EEProm.nodeID & 0x1F);
          break;

        case 'b':  // set band: 4 = 433, 8 = 868, 9 = 915
          EEProm.RF_freq = bandToFreq(Serial.parseFloat());
          Serial.print(EEProm.RF_freq == RF12_433MHZ ? 433 : 
                       EEProm.RF_freq == RF12_868MHZ ? 868 :
                       EEProm.RF_freq == RF12_915MHZ ? 915 : 0);
          Serial.println(F(" MHz"));
          break;

        case 'f':
          /*
          *  Format expected: f50 | f60
          */
          k1 = Serial.parseFloat();
          EmonLibCM_cycles_per_second(k1);
          Serial.print(F("Freq: "));Serial.println(k1);
          break;
        
        case 'g':  // set network group
          EEProm.networkGroup = Serial.parseFloat();
          Serial.print(F("Group ")); Serial.println(EEProm.networkGroup);
          break;

        case 'k': // On-line calibration - required due to faulty ESP8266 software
          calibration_enable = true; 
          Serial.println(F("On-line calibration is now ENABLED"));
          break;

        case 'l': // print the calibration values
          list_calibration();
          break;

        case 'p': // set RF power level
          EEProm.rf_power = (Serial.parseInt() & 0x1F) | 0x80;
          Serial.print("p = ");Serial.print((EEProm.rf_power & 0x1F) - 18);Serial.println(" dBm");
          break; 
          
        case 'r': // restore sketch defaults
          wipe_eeprom();
          softReset();
          break;

        case 's': // Save to EEPROM. ATMega328p has 1kB  EEPROM
          save_config();
          break;

        case 'v': // print firmware version
          Serial.print(F("emonTx V3.4 + RFM69CW EmonLibCM Continuous Monitoring Special V")); Serial.println(version*0.1);
          break;
        
        case 'w' :  // RF Off / On / On & whitened
          /* Format expected: w[x]
           */
          EEProm.rf_whitening = Serial.parseFloat(); 
          Serial.print(EEProm.rf_whitening ? (EEProm.rf_whitening ==1 ? "RF on":"RF whitened"):"RF off"); Serial.print("\n");
          break;
          
        case 'x':  // exit and continue
          Serial.print(F("\nContinuing...\n"));
          while (Serial.available())
            Serial.read(); 
          return false;
          break;

        case '?':  // show Help text        
          showString(helpText1);
          Serial.println(F(" "));
          break;
        
        default:
          break;
    } //end switch
    while (Serial.available())
      Serial.read(); 
  } 
  return true;
}



void getCalibration(void)
{
/*
 * Reads calibration information (if available) from the serial port at runtime. 
 * Data is expected generally in the format
 * 
 *  [l] [x] [y] [z]
 * 
 * where:
 *  [l] = a single letter denoting the variable to adjust
 *  [x] [y] [z] are values to be set.
 *  see the user instruction above, the comments below or the separate documentation for details
 * 
 */

	if (Serial.available() && calibration_enable)
  {   
    int k1; 
    double k2, k3; 
    char c = Serial.peek();
    switch (c) {
      
      case 'c':
        /*
        * Format expected: c0 | c1
        */
        k1 = Serial.parseInt();
        switch (k1) {
          case 1 : EEProm.showCurrents = true;
            break;
          default: EEProm.showCurrents = false;
        }
        break;
      
      case 'k':
        /*  Format expected: k[x] [y] [z]
        * 
        * where:
        *  [x] = a single numeral: 0 = voltage calibration, 1 = ct1 calibration, 2 = ct2 calibration, etc
        *  [y] = a floating point number for the voltage/current calibration constant
        *  [z] = a floating point number for the phase calibration for this c.t. (z is not needed, or ignored if supplied, when x = 0)
        * 
        * e.g. k0 256.8
        *      k1 90.9 1.7 
        * 
        * If power factor is not displayed, it is impossible to calibrate for phase errors,
        *  and the standard value of phase calibration MUST BE SENT when a current calibration is changed.
        */
        k1 = Serial.parseFloat(); 
        k2 = Serial.parseFloat(); 
        k3 = Serial.parseFloat(); 
        while (Serial.available())
          Serial.read(); 
              
        // Write the values back as Globals, re-calculate intermediate values.
        switch (k1) {
          case 0 : EmonLibCM_ReCalibrate_VChannel(k2);
            if (USA)
              EEProm.vCal_USA = k2;
            else
              EEProm.vCal = k2;
            break;
              
          case 1 : EmonLibCM_ReCalibrate_IChannel(1, k2, k3);
            EEProm.i1Cal = k2;
            EEProm.i1Lead = k3;
            break;

          case 2 : EmonLibCM_ReCalibrate_IChannel(2, k2, k3);
            EEProm.i2Cal = k2;
            EEProm.i2Lead = k3;
            break;

          case 3 : EmonLibCM_ReCalibrate_IChannel(3, k2, k3);
            EEProm.i3Cal = k2;
            EEProm.i3Lead = k3;
            break;

          case 4 : EmonLibCM_ReCalibrate_IChannel(4, k2, k3);
            EEProm.i4Cal = k2;
            EEProm.i4Lead = k3;
            break;
                       
          default : ;
        }
        Serial.print(F("Cal: k"));Serial.print(k1);Serial.print(F(" "));Serial.print(k2);Serial.print(F(" "));Serial.println(k3);        
        break;
          
      case 'l':
        list_calibration(); // print the calibration values
      break;
          
      case 'm' :
        /*  Format expected: m[x] [y]
         * 
         * where:
         *  [x] = a single numeral: 0 = pulses OFF, 1 = pulses ON,
         *  [y] = an integer for the pulse min period in ms - ignored when x=0
         */
        k1 = Serial.parseFloat(); 
        k2 = Serial.parseFloat(); 
        while (Serial.available())
          Serial.read(); 

        switch (k1) {
          case 0 : EmonLibCM_setPulseEnable(false);
            EEProm.pulse_enable = false;
            break;
          
          case 1 : EmonLibCM_setPulseMinPeriod(k2);
            EmonLibCM_setPulseEnable(true);
            EEProm.pulse_enable = true;
            EEProm.pulse_period = k2;
            break;
        }
        Serial.print(F("Pulses: "));
        if (k1)
          {Serial.print(k2);Serial.println(F(" ms"));}
        else
          Serial.println("off");        
        break;        
  
      case 'p':
        /*  Format expected: p[x]
         * 
         * where:
         *  [x] = a floating point number for the datalogging period in s
         */
        k2 = Serial.parseFloat(); 
        EmonLibCM_datalog_period(k2); 
        EEProm.period = k2;
        Serial.print(F("datalog period: ")); Serial.print(k2);Serial.println(F(" s"));
        break;

      case 's' :
        save_config(); // Save to EEPROM. ATMega328p has 1kB  EEPROM
        break;
          
      case 't' :
        /*  Format expected: t[x] [y] [y] ...
         */
        set_temperatures();
        break;

      case '?':  // show Help text        
        showString(helpText1);
        Serial.println(F(" "));
        break;

    }
    // flush the input buffer
    while (Serial.available())
      Serial.read();
  }
}


static byte bandToFreq (byte band) {
  return band == 4 ? RF12_433MHZ : band == 8 ? RF12_868MHZ : band == 9 ? RF12_915MHZ : 0;
}


static void showString (PGM_P s) {
  for (;;) {
    char c = pgm_read_byte(s++);
    if (c == 0)
      break;
    if (c == '\n')
      Serial.print('\r');
    Serial.print(c);
  }
}

void set_temperatures(void)
{
  /*  Format expected: t[x] [y] [y] ...
  * 
  * where:
  *  [x] = 0  [y] = single numeral: 0 = temperature measurement OFF, 1 = temperature measurement ON
  *  [x] = a single numeral > 0: the position of the sensor in the list (1-based)
  *  [y] = 8 hexadecimal bytes representing the sensor's address
  *          e.g. t2 28 81 43 31 07 00 00 D9
  */
    
  DeviceAddress sensorAddress;
         
	int k1 = Serial.parseFloat();

  if (k1 == 0)
  { 
    // write to EEPROM
    EEProm.temp_enable = Serial.parseInt();
    EmonLibCM_TemperatureEnable(EEProm.temp_enable);
  }
  else if (k1 > EmonLibCM_getTemperatureSensorCount())
    return;
  else
  {
    byte i = 0, a, b;
    Serial.readBytes(&b,1);     // expect a leading space
    while (Serial.readBytes(&b,1) && i < 8)
    {            
      if (b == ' ' || b == '\r' || b == '\n')
      {
        sensorAddress[i++] = a;
        a = 0;
      }                
      else
      {
        a *= 16;
        a += c2h(b);
      }          
    }     
    // set address
    for (byte i=0; i<8; i++)
    {
      allAddresses[k1-1][i] = sensorAddress[i];
      EEProm.temp_addr[(k1-1)*8+i] = sensorAddress[i];
    }
  }
	while (Serial.available())
		Serial.read(); 
}

byte c2h(byte b)
{
  if (b > 47 && b < 58) 
    return b - 48;
  else if (b > 64 && b < 71) 
    return b - 55;
  else if (b > 96 && b < 103) 
    return b - 87;
  return 0;
}
