/*
 * emonEEProm.cpp
 *
 * Library to manage configuration data stored in EEPROM
 *
 */
 
#include <Arduino.h>
#include <avr/pgmspace.h>
#include <EEPROM.h>
#include "emonEEProm.h"

char moniker[3] = {'O', 'E', 'M'};

unsigned int eepromSize(void)
{
/*
 * Get EEPROM size (duplicates EEPROM.length() )
 * Returns total size of EEPROM (E2END is from Arduino header file)
 */
 
  return E2END + 1;
}


unsigned int eepromFree(void)
{
/*
 * Returns size of EEPROM free space, or zero if not valid OEM format
 */ 
 
  block b;
  unsigned int offset = sizeof(moniker);
  if (!eepromValidate())
    return 0;

  do
  {
    EEPROM.get(offset, b);
    if (b.signature == 0xFFFF && b.length == 0xFFFF)
      return (E2END + 1 - offset);
    else
      offset += abs(b.length);
  } while (offset <= E2END);
  return 0;
}


void eepromFormat(void)
{
/*
 * Unconditionally erases all EEPROM contents and
 * writes OEM signature.
 */
 
  for (byte j=0; j<sizeof(moniker); j++)
    EEPROM.update(j, moniker[j]);
  for (unsigned int j=sizeof(moniker); j<=E2END; j++)
    EEPROM.update(j, 0xFF);    
}

  
bool eepromValidate(void)
{
/*
 * Checks for the existence of the "OEM" signature.
 * Returns TRUE if found.
 */
 
  for (byte j=0; j<sizeof(moniker); j++)
    if (EEPROM.read(j) != moniker[j])
      return false;
    return true;
}


unsigned int eepromFindSignature(uint16_t signature) 
{
/*
 * Find the signature and return the offset from the beginning of EEPROM. 
 * Returns zero if not found.
 */
 
  unsigned int offset = sizeof(moniker);
  
  if (!eepromValidate)
    return 0;
  
  do
  {
    block b; 
    EEPROM.get(offset, b);
    if (b.signature == signature)
      return offset;
    else 
    if (b.signature == 0xFFFF && b.length == 0xFFFF)
      return 0;
    else
      offset += abs(b.length);
  } while (offset <= E2END);
  return 0;
}


unsigned int eepromRead(uint16_t signature, byte *dest)
{
/*
 * Find the block identified by signature and copies the EEPROM data to the struct "dest", which MUST be big enough. 
 * Returns zero if not found, or EEPROM block has been hidden.
 */
 
  unsigned int offset = eepromFindSignature(signature);
  if (offset)
  {
    block b;
    EEPROM.get(offset, b);
    if (b.length < 0)
      return 0;
    unsigned int end = offset + b.length;
    offset += sizeof(b);
    
    do {
      *dest++ = EEPROM.read(offset++);
    } while (offset < end);
    return b.length-sizeof(b);  
  }
  return 0;
}


unsigned int eepromWrite(uint16_t signature, byte *src, int16_t length)
{
/*
 * Writes unconditionally to unrecognised EEPROM, else finds the block identified by signature,
 * else finds the end, and then copies the struct "src" data of length 'length' bytes to the EEPROM. 
 * Returns the number of bytes written (=length + block header), or zero if unable to write. 
 */

  unsigned int offset = sizeof(moniker);
  if (!eepromValidate())                                        // EEPROM format unrecognised - use it
  {   
    Serial.println("New EEPROM - formatting...");
    eepromFormat();
  }
  else                                                          // Search for the signature
  {
    block b;
    if ((offset = eepromFindSignature(signature)) == 0)         // find the end instead
    {
      offset+= sizeof(moniker);
      do
      {
        EEPROM.get(offset, b);
        if (b.signature == 0xFFFF && b.length == 0xFFFF)
          break;
        else
          offset += abs(b.length);
      } while (offset <= E2END);
      if (offset + length > E2END)
        return 0;                                               // Not enough space remains
    }                                                           // Found the signature or the end
    if (length > (E2END - offset - sizeof(moniker) - sizeof(b)))// Asking to write beyond EEPROM block
      return 0;
  }                                                             // EEPROM is newly formatted
  
  // Write the block
  block b;
  b.length = length + sizeof(b);
  int end = offset + b.length;
  b.signature = signature;
  EEPROM.put(offset, b);
  offset += sizeof(b);
  do {
    EEPROM.update(offset++, *src++);
  } while (offset < end && offset < E2END);
  return b.length;
}


unsigned int eepromHide(uint16_t signature)
{
/*
 * Finds and marks the block identified by signature as hidden. 
 * Returns the number of data bytes in the block (=length), or zero if unable to find the signature. 
 */

  block b;
  unsigned int offset;
  if (!eepromValidate())                                        // EEPROM format unrecognised
    return 0;

  if ((offset = eepromFindSignature(signature)) == 0)
    return 0;                                                   // Signature not found
    
  EEPROM.get(offset, b);
  if (b.length < 0)
    return 0;                                                   // Already hidden!
  int end = offset + b.length;
  b.length = -b.length;
  EEPROM.put(offset, b);
  return -b.length;
}


unsigned int eepromLift(uint16_t signature, int16_t length)
{
/*
 * Lifts an existing EEPROM configuration (i.e. NOT formatted according to this standard) and
 * rewrites it to comply.
 *
 * "signature" is the NEW signature for the data, "length" is the length of the existing data.
 */
 
  block b;
  unsigned int offset, src, dest;
  if (eepromValidate())                                        // New EEPROM format recognised - do nothing
    return 0;  
  for (src = length, dest = length + sizeof(moniker) + sizeof(b); src > 0; )
    EEPROM.update(--dest, EEPROM.read(--src));
  EEPROM.put(0, moniker);
  b.length = length + sizeof(b);
  b.signature = signature;
  EEPROM.put(sizeof(moniker), b);
  return (length + sizeof(moniker) + sizeof(b));
}


unsigned int eepromPack(void)
{
/*
 * Finds blocks marked as hidden, moves subsequent blocks down to fill the vacated space.
 * returns the total number of used bytes, or zero if the format is not recognised and nothing is done.
 */

  block a, b;
  unsigned int offset, end, nextStart;
  if (!eepromValidate())                                       // EEPROM format unrecognised - do nothing.
  {
   return 0;
  }
  offset = sizeof(moniker);
  EEPROM.get(offset, a);
  if (a.length == 0xFFFF)
  {
    return 0;
  }
  while ((offset + abs(a.length)) < (E2END - sizeof(a)) && a.length !=0xFFFF)
  {
    // Find first unused block or the end
    while (a.signature != 0xFFFF && a.length > 0 && offset < (E2END - sizeof(a)))
    {
      offset += a.length;
      EEPROM.get(offset, a);
    }
    if ((a.signature == 0xFFFF && a.length == 0xFFFF) || offset >= (E2END - sizeof(a)))           
                                                                // no more blocks, nothing to do.
      return offset;
    // found a hidden block, get the next into b.
    EEPROM.get(offset+abs(a.length), b);
    if (b.length == 0xFFFF)                                     // the next is the end
    {
      for (int j = offset; j<E2END; j++)
        EEPROM.update(j, 0xFF);
      return offset;
    }

    if (b.length < 0)                                           // or the next block has been hidden
    {
      a.length += b.length;                                     // ADD lengths (both negative)
      EEPROM.put(offset, a);                                    // and the two blocks are merged 
      continue;
    }
    else                                                        // else the next block is in use
    {
      unsigned int src, dest, length;
      for (length = b.length, src = offset-a.length, dest = offset; length > 0; src++, dest++, length--) // a.length is negative!
        EEPROM.update(dest, EEPROM.read(src));
        
      a.signature = 0xFFFF;
      offset += b.length;
      EEPROM.put(offset, a);                                    // mark the newly vacated area as a hidden block 
    }
  }
  return offset;                                                // Should never return here.
}

void eepromPrint(void)
{
/*
 * Print to Serial the entire used contents of EEPROM.
 * If the format is unrecognised, the complete contents of EEPROM is printed, else each block is printed individually.
 * 
 * Assumes Serial printing is available via host sketch.
 */
  block b;
  unsigned int offset;
  if (!eepromValidate())                                        // EEPROM format not recognised - print raw.
  {
    Serial.print("Size: ");Serial.print(eepromSize());Serial.println(" bytes, not formatted to OEM Standard");
    printblock(0, eepromSize());
    Serial.print("\n");
  }
  else                                                          // OEM-formatted - print by block.
  {
    offset = sizeof(moniker);
// for each block  

    int j = 0;
    EEPROM.get(offset, b);
    do
    {
      if ((j == 0 && b.signature == 0xFFFF))
      {
        Serial.println("No data blocks found - EEPROM appears to be formatted to OEM Standard but unused.");
        return;
      }
      Serial.print("\nBlock "); Serial.print(++j);Serial.print(" Signature: ");Serial.print(b.signature);
      int size = abs(b.length) - sizeof(block);
      Serial.print(", Data size: ");Serial.print(size);Serial.print(" bytes");
      if (b.length < 0)
          Serial.print(" (hidden).");
      Serial.println("");  
      printblock(offset+sizeof(block), size);
      Serial.print("\n");
      offset += (size + sizeof(block));
      EEPROM.get(offset, b);   
    }
    while (b.length != 0xFF && b.signature != 0xFFFF && (offset + abs(b.length)) < E2END);
    Serial.println("\nNo more data blocks found.");
  }
}


bool eepromCheck(uint16_t signature, byte *src, int16_t length)
{
/*
 * Find the block identified by signature and checks the EEPROM data against the struct "src". 
 * Returns true if the EEPROM block matches in content and length, else false. 
 * (Also works for a block flagged as 'hidden'.)
 */
 
  unsigned int offset = eepromFindSignature(signature);
  if (offset)
  {
    block b;
    EEPROM.get(offset, b);
    int size = abs(b.length);

    if (size == 0)
      return false;
          
    if (length != b.length - sizeof(b))
      return false;
          
    for (int j = offset + sizeof(b); j < offset + size;)
    {
      if (*src++ != EEPROM.read(j++))
        return false;
    }
    return true;  
  }
  return false;
}


void printblock(int start, int size)
{
  // private to the library
  for (int j=0; j<size;)
  {
    if (EEPROM[start + j] < 0x10)
      Serial.print('0');
    Serial.print(EEPROM[start + j++],16);Serial.print(" ");

    if (j % 8 == 0)
      Serial.print(" ");

    if (j % 16 == 0)
      Serial.print("\n");    
  }
}