/*
  emonTxV3.2 Continuous Sampling
  using EmonLibCM https://github.com/openenergymonitor/EmonLibCM
  Authors: Robin Emley, Robert Wall, Trystan Lea
  
  -----------------------------------------
  Part of the openenergymonitor.org project
  Licence: GNU GPL V3
*/

/*
Change Log:
v0.1: First version, special for porboynz (Kevin Ross)
      combination of EmonTxV3CM_max and EmonTx v3 original firmware

emonhub.conf node decoder
See: https://github.com/openenergymonitor/emonhub/blob/emon-pi/configuration.md
copy the following in to emonhub.conf:

[[15]]
  nodename = emontx15
  [[[rx]]]
    names = MSG, Vrms, P1, P2, P3, P4, E1, E2, E3, E4, T1, T2, T3, pulse
    datacodes = L,h,h,h,h,h,l,l,l,l,h,h,h,L
    scales = 1,0.01,1,1,1,1,1,1,1,1,0.01,0.01,0.01,1
    units = n,V,W,W,W,W,Wh,Wh,Wh,Wh,C,C,C,p
    whitening = 1

*/
#include <Arduino.h>

const byte version = 1;                                 // Firmware version divide by 10 to get version number e,g 05 = v0.5

// Comment/Uncomment as applicable
#define DEBUG                                        // Debug level print out

#define RFM69CW
#define RFMSELPIN 4                                     // RFM pins
#define RFMIRQPIN 3                                     // RFM pins

#include "emonLibCM.h"

#include <Wire.h>                                       // Required for RFM & temperature measurement
#include <SPI.h>
#include <util/crc16.h>
#include <OneWire.h>
#include <emonEEProm.h>

enum rfband {RF12_433MHZ = 1, RF12_868MHZ, RF12_915MHZ }; // frequency band.

const int busyThreshold = -97;                          // Signal level below which the radio channel is clear to transmit
const byte busyTimeout = 15;                            // Time in ms to wait for the channel to become clear, before transmitting anyway

/*
typedef struct {
    unsigned long Msg;
    int Vrms,P1,P2,P3,P4; 
    long E1,E2,E3,E4; 
    int T1,T2,T3;
    unsigned long pulse;
} PayloadTX;
*/
typedef struct {
    int P1,P2,P3,P4,Vrms; 
    int T1, T2;
} PayloadTX;
PayloadTX emontx;                                       // create an instance
static void showString (PGM_P s);
 
DeviceAddress allAddresses[3];                          // Array to receive temperature sensor addresses
/*   Example - how to define temperature sensors, prevents an automatic search
DeviceAddress allAddresses[] = {       
    {0x28, 0x81, 0x43, 0x31, 0x7, 0x0, 0xFF, 0xD9}, 
    {0x28, 0x8D, 0xA5, 0xC7, 0x5, 0x0, 0x0, 0xD5},      // Use the actual addresses, as many as required
    {0x28, 0xC9, 0x58, 0x32, 0x7, 0x0, 0x0, 0x89}       // up to a maximum of 6    
};
*/

int allTemps[3];                                        // Array to receive temperature measurements

//----------------------------emonTx V3 Settings - Stored in EEPROM and shared with config.ino------------------------
struct {
  byte nodeID = 10;                                     // node ID for this emonTx.
  byte RF_freq = RF12_433MHZ;                           // Frequency of radio module can be RF12_433MHZ, RF12_868MHZ or RF12_915MHZ. 
  int networkGroup = 210;                               // wireless network group, needs to be same as emonBase / emonPi and emonGLCD. OEM default is 210
  float vCal  = 320.4;                                  // (240V x 13) / 11.6V = 268.97 Calibration for UK AC-AC adapter 77DB-06-09
  float i1Cal = 90.9;                                   // (100 A / 50 mA / 22 Ohm burden) = 90.9
  float i1Lead = 4.2;
  float i2Cal = 90.9;                                   // (100 A / 50 mA / 22 Ohm burden) = 90.9
  float i2Lead = 4.2;
  float i3Cal = 90.9;                                   // (100 A / 50 mA / 22 Ohm burden) = 90.9
  float i3Lead = 4.2;
  float i4Cal = 16.67;                                  // (100 A / 50 mA / 120 Ohm burden) = 16.67
  float i4Lead = 1.0;
  float period = 9.96;                                  // datalogging period
  bool  pulse_enable = true;                            // pulse counting
  int   pulse_period = 100;                             // pulse min period
  bool  temp_enable = true;                             // enable temperature measurement
  byte  temp_addr[24];                                  // sensor address data
  int   rf_whitening = 1;                               // RF & data whitening - 0 = no RF, 1 = RF on, no whitening, default = 2: RF is ON with whitening.
  byte  rf_power = 0x99;                                // 0x80 = -18 dBm, 0x9F = +13 dBm, default = 0x99\n"
  bool  showCurrents = false;                           // Print to serial current & p.f. values
  float vCal_USA = 130.0;                               // Calibration for US AC-AC adapter 77DA-10-09
  
} EEProm;

uint16_t eepromSig = 0xFF01;
bool USA = false;
bool calibration_enable = false;                        // Enable on-line calibration when running.

//----------------------------emonTx V3 hard-wired connections-----------------------------------
const byte LEDpin      = 6;  // emonTx V3.2 LED
const byte DS18B20_PWR = 19; // DS18B20 Power

//---------------------------------CT availability status----------------------------------------
byte CT_count = 0;
bool CT1, CT2, CT3, CT4;     // Record if CT present during startup


//----------------------------------------Setup--------------------------------------------------
void setup() 
{  
  pinMode(LEDpin, OUTPUT);
  digitalWrite(LEDpin,HIGH);
  
  pinMode(DS18B20_PWR, OUTPUT);
  digitalWrite(DS18B20_PWR, HIGH);
  
  // Serial---------------------------------------------------------------------------------
  Serial.begin(115200);

  // ---------------------------------------------------------------------------------------
  #ifdef DEBUG
    Serial.print(F("emonTx V3.2 EmonLibCM Continuous Monitoring Special - V")); Serial.println(version*0.1);
    Serial.println(F("OpenEnergyMonitor.org"));
  #else
    Serial.println(F("describe:EmonTX3CM"));
  #endif
 
  load_config(true);                                                        // Load RF config from EEPROM (if any exists)
  
  if (EEProm.rf_whitening)
  {
    #ifdef DEBUG
      Serial.print(F(" Node: ")); Serial.print(EEProm.nodeID);
      Serial.print(" Freq: ");
      if (EEProm.RF_freq == RF12_433MHZ) Serial.print(F("433MHz"));
      if (EEProm.RF_freq == RF12_868MHZ) Serial.print(F("868MHz"));
      if (EEProm.RF_freq == RF12_915MHZ) Serial.print(F("915MHz"));
      Serial.print(F(" Group: ")); Serial.println(EEProm.networkGroup);
      Serial.println(" ");
    #endif
  }
  
  
  
  // Check connected CT sensors ------------------------------------------------------------
  if (analogRead(1) > 0) {CT1 = 1; CT_count++;} else CT1=0;               // check to see if CT is connected to CT1 input
  if (analogRead(2) > 0) {CT2 = 1; CT_count++;} else CT2=0;               // check to see if CT is connected to CT2 input
  if (analogRead(3) > 0) {CT3 = 1; CT_count++;} else CT3=0;               // check to see if CT is connected to CT3 input
  if (analogRead(4) > 0) {CT4 = 1; CT_count++;} else CT4=0;               // check to see if CT is connected to CT4 input

  // ---------------------------------------------------------------------------------------
  readConfigInput();
  if (EEProm.rf_whitening)
  {
    rfm_init(EEProm.RF_freq);                                             // initialize RFM
    for (int i=10; i>=0; i--)                                             // Send RF test sequence (for factory testing)
    {
      emontx.P1=i;
      rfm_send((byte *)&emontx, sizeof(emontx), EEProm.networkGroup, EEProm.nodeID);
      delay(100);
    }
    emontx.P1=0;
  }
  
  // ---------------------------------------------------------------------------------------
  
  digitalWrite(LEDpin,LOW);

  #ifdef DEBUG
    if (CT_count==0) {
      Serial.println(F("NO CT's detected"));
    } else {
      if (CT1) { Serial.print(F("CT1 detected, i1Cal:")); Serial.println(EEProm.i1Cal); }
      if (CT2) { Serial.print(F("CT2 detected, i2Cal:")); Serial.println(EEProm.i2Cal); }
      if (CT3) { Serial.print(F("CT3 detected, i3Cal:")); Serial.println(EEProm.i3Cal); }
      if (CT4) { Serial.print(F("CT4 detected, i4Cal:")); Serial.println(EEProm.i4Cal); }
    }
    delay(200);
  #endif

  // ----------------------------------------------------------------------------
  // EmonLibCM config
  // ----------------------------------------------------------------------------
  EmonLibCM_SetADC_VChannel(0, EEProm.vCal);                      // ADC Input channel, voltage calibration - for Ideal UK Adapter = 268.97
  if (USA) EmonLibCM_SetADC_VChannel(0, EEProm.vCal_USA);         // ADC Input channel, voltage calibration - for US AC-AC adapter 77DA-10-09
  EmonLibCM_SetADC_IChannel(1, EEProm.i1Cal, EEProm.i1Lead);      // ADC Input channel, current calibration, phase calibration
  EmonLibCM_SetADC_IChannel(2, EEProm.i2Cal, EEProm.i2Lead);      // The current channels will be read in this order
  EmonLibCM_SetADC_IChannel(3, EEProm.i3Cal, EEProm.i3Lead);      // 90.91 for 100 A : 50 mA c.t. with 22R burden - v.t. leads c.t by ~4.2 degrees
  EmonLibCM_SetADC_IChannel(4, EEProm.i4Cal, EEProm.i4Lead);      // 16.67 for 100 A : 50 mA c.t. with 120R burden - v.t. leads c.t by ~1 degree

  EmonLibCM_ADCCal(3.3);                                   // ADC Reference voltage, (3.3 V for emonTx,  5.0 V for Arduino)
  // mains frequency 50Hz
  if (USA) EmonLibCM_cycles_per_second(60);                // mains frequency 60Hz
  EmonLibCM_datalog_period(EEProm.period);                 // period of readings in seconds - normal value for emoncms.org  

  EmonLibCM_setPulseEnable(EEProm.pulse_enable);           // Enable pulse counting
  EmonLibCM_setPulsePin(2, 0);
  EmonLibCM_setPulseMinPeriod(EEProm.pulse_period);

  EmonLibCM_setTemperatureDataPin(5);                      // OneWire data pin (emonTx V3.2 / V3.4)
  EmonLibCM_setTemperaturePowerPin(19);                    // Temperature sensor Power Pin - 19 for emonTx V3.2 / V3.4  (-1 = Not used. No sensors, or sensor are permanently powered.)
  EmonLibCM_setTemperatureResolution(11);                  // Resolution in bits, allowed values 9 - 12. 11-bit resolution, reads to 0.125 degC
  EmonLibCM_setTemperatureAddresses(allAddresses);         // Name of array of temperature sensors
  EmonLibCM_setTemperatureArray(allTemps);                 // Name of array to receive temperature measurements
  EmonLibCM_setTemperatureMaxCount(2);                     // Max number of sensors, limited by wiring and array size.
  
  EmonLibCM_TemperatureEnable(EEProm.temp_enable);  
  EmonLibCM_Init();                                        // Start continuous monitoring.
  
  //emontx.Msg = 0;
  
}

void loop()             
{
  getCalibration();
  
  if (EmonLibCM_Ready())   
  {
    /*
    #ifdef DEBUG
    if (emontx.Msg==0) {
      Serial.println(EmonLibCM_acPresent()?F("AC present "):F("AC missing "));
      delay(5);
    }
    #endif

    emontx.Msg++;
    */

    // Other values calculated by EmonLibCM
    // RMS Current:    EmonLibCM_getIrms(ch)
    // Apparent Power: EmonLibCM_getApparentPower(ch)
    // Power Factor:   EmonLibCM_getPF(ch)
    
    emontx.P1 = EmonLibCM_getRealPower(0); 
    //emontx.E1 = EmonLibCM_getWattHour(0); 

    emontx.P2 = EmonLibCM_getRealPower(1); 
    //emontx.E2 = EmonLibCM_getWattHour(1); 
    
    emontx.P3 = EmonLibCM_getRealPower(2); 
    //emontx.E3 = EmonLibCM_getWattHour(2); 
  
    emontx.P4 = EmonLibCM_getRealPower(3); 
    //emontx.E4 = EmonLibCM_getWattHour(3); 

    emontx.Vrms = EmonLibCM_getVrms() * 100;
    
    emontx.T1 = allTemps[0];
    emontx.T2 = allTemps[1];
    //emontx.T3 = allTemps[2];

    //emontx.pulse = EmonLibCM_getPulseCount();
    
    if (EEProm.rf_whitening)
    {
      PayloadTX tmp = emontx;
      if (EEProm.rf_whitening == 2)
      {
          byte WHITENING = 0x55;
          for (byte i = 0, *p = (byte *)&tmp; i < sizeof tmp; i++, p++)
              *p ^= (byte)WHITENING;
      }
      rfm_send((byte *)&tmp, sizeof(tmp), EEProm.networkGroup, EEProm.nodeID); //send data
      delay(50);
    }

    // ---------------------------------------------------------------------
    // Key:Value format, used by EmonESP & emonhub EmonHubTx3eInterfacer
    // ---------------------------------------------------------------------
    else
    {
      //Serial.print(F("MSG:")); Serial.print(emontx.Msg);
      Serial.print(F("Vrms:")); Serial.print(emontx.Vrms*0.01);
      
      if (CT1) { Serial.print(F(",P1:")); Serial.print(emontx.P1); }
      if (CT2) { Serial.print(F(",P2:")); Serial.print(emontx.P2); }
      if (CT3) { Serial.print(F(",P3:")); Serial.print(emontx.P3); }
      if (CT4) { Serial.print(F(",P4:")); Serial.print(emontx.P4); }
      
      //if (CT1) { Serial.print(F(",E1:")); Serial.print(emontx.E1); }
      //if (CT2) { Serial.print(F(",E2:")); Serial.print(emontx.E2); }
      //if (CT3) { Serial.print(F(",E3:")); Serial.print(emontx.E3); }
      //if (CT4) { Serial.print(F(",E4:")); Serial.print(emontx.E4); }
      
      if (emontx.T1!=30000) { Serial.print(F(",T1:")); Serial.print(emontx.T1*0.01); }
      if (emontx.T2!=30000) { Serial.print(F(",T2:")); Serial.print(emontx.T2*0.01); }
      //if (emontx.T3!=30000) { Serial.print(F(",T3:")); Serial.print(emontx.T3*0.01); }
  
      //Serial.print(F(",pulse:")); Serial.println(emontx.pulse);  
      delay(20);
    }
    digitalWrite(LEDpin,HIGH); delay(50);digitalWrite(LEDpin,LOW);

    if (EEProm.showCurrents)
    {
      // to show current & power factor for calibration:
  
      Serial.print(F("Vrms:")); Serial.print(emontx.Vrms*0.01);
      Serial.print(F(",I1:")); Serial.print(EmonLibCM_getIrms(EmonLibCM_getLogicalChannel(1)));
      Serial.print(F(",I2:")); Serial.print(EmonLibCM_getIrms(EmonLibCM_getLogicalChannel(2)));
      Serial.print(F(",I3:")); Serial.print(EmonLibCM_getIrms(EmonLibCM_getLogicalChannel(3)));
      Serial.print(F(",I4:")); Serial.print(EmonLibCM_getIrms(EmonLibCM_getLogicalChannel(4)));

      Serial.print(F(",pf1:")); Serial.print(EmonLibCM_getPF(EmonLibCM_getLogicalChannel(1)),4);
      Serial.print(F(",pf2:")); Serial.print(EmonLibCM_getPF(EmonLibCM_getLogicalChannel(2)),4);
      Serial.print(F(",pf3:")); Serial.print(EmonLibCM_getPF(EmonLibCM_getLogicalChannel(3)),4);
      Serial.print(F(",pf4:")); Serial.print(EmonLibCM_getPF(EmonLibCM_getLogicalChannel(4)),4);
      Serial.print(F("\n"));
    }
    // End of print out ----------------------------------------------------
  }
}
